#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "cluster.h"

void unbuffer_terminal(struct termios *save)
{
	struct termios termios;
	if (tcgetattr(0, &termios) < 0)
		goto error;
	*save = termios;
	termios.c_iflag |= IUCLC;
	termios.c_lflag &= ~(ICANON|ECHO);
	if (tcsetattr(0, TCSANOW, &termios) < 0)
		goto error;
	return;
error:
	printf("Error setting stdin to unbuffered mode: %s\n", strerror(errno));
	exit(1);
}

void restore_terminal(struct termios *save)
{
	tcsetattr(0, TCSANOW, save);
}

int
main(int argc, char **argv)
{
	struct termios save;
	struct stat statbuf;
	clusternode_t *nodes;
	int
		fd = 3, 
		pipefds[2] = {4, 5}, 
		ret = 1, 
		maxnodes = cluster_maxnodes(),
		randctr = 0;
	char 
		buf, 
		random = 0;

	nodes = malloc(maxnodes * sizeof(*nodes));
	if (!nodes)
		goto err_nomem;

	unbuffer_terminal(&save);

	if (fstat(pipefds[0], &statbuf) == 0)
		random = 1;

	if (fstat(fd, &statbuf) == 0) {
		printf("Moved to node %d\n", clusternode_num());
		goto re_exec;
	}

	if (argc != 2)
		goto err_nofile;
	fd = open(argv[1], O_RDONLY);
	if (fd != 3)
		goto err_nofile;

	if (fcntl(0, F_SETFL, O_NONBLOCK) < 0)
		goto err_fcntl;

re_exec:
	for (;;) {
		ssize_t count;
		time_t curtime;

		for (;;) {
			count = read(fd, &buf, 1);
			if (count < 0)
				goto err_read;
			if (!count) {
				lseek(fd, 0, SEEK_SET);
				continue;
			}
			printf("%.1s", &buf);
			if (buf == '\n')
				break;
		}

		curtime = time(NULL);
		for (;;) {
			if (time(NULL) != curtime)
				break;

			count = read(0, &buf, 1);
			if (count < 0)
				if (errno == EAGAIN)
					continue;
				else
					goto err_read;

			if (buf == '\n') {
				if (random)
					continue;
				random = 1;
				if (pipe(pipefds) < 0 || pipefds[0] != 4)
					goto err_pipe;
				close(pipefds[1]);
			} else if (random) {
				random = 0;
				close(pipefds[0]);
			}
			if (buf >= '1' && buf <= '9') {
				clusternode_t node = (clusternode_t)(buf - '0');
				if (clusternode_avail(node) != 1)
					continue;
				restore_terminal(&save);
				if (rexecvp(argv[0], argv, node) < 0)
					goto err_rexec;
			}
			if (buf == 'q')
				goto quit;
		}

		if (random)
			++randctr;
		if (randctr == 10) {
			int nodecount;
			clusternode_t node;
			randctr = 0;

			nodecount = cluster_membership(NULL, maxnodes, nodes);
			if (nodecount < 0)
				goto err_member;
			srand(time(NULL));
			node = nodes[(int)(nodecount * (rand() + 0.0) 
						/ RAND_MAX)];
			restore_terminal(&save);
			if (rexecvp(argv[0], argv, node) < 0)
				goto err_rexec;
		}
	}

quit:
	ret = 0;
out:
	restore_terminal(&save);
	return ret;
err_nomem:
	printf("Out of memory");
	return 1;
err_nofile:
	printf("Must specify an existing filename\n");
	goto out;
err_fcntl:
	printf("Error setting stdin to non-blocking: %s\n", strerror(errno));
	goto out;
err_read:
	printf("Error reading from file: %s\n", strerror(errno));
	goto out;
err_pipe:
	printf("Error creating pipe: %s\n", strerror(errno));
	goto out;
err_member:
	printf("Error getting membership info: %s\n", strerror(errno));
	goto out;
err_rexec:
	printf("Error rexec'ing: %s\n", strerror(errno));
	goto out;
}
